/* Code to save the arptables state, in human readable-form. */
/* Written by Jay Fenlason <fenlason@redhat.com>
 * Copyright 2003 Red Hat
 * Based on iptables-save.c by */
/* (C) 1999 by Paul 'Rusty' Russell <rusty@rustcorp.com.au> and
 * (C) 2000-2002 by Harald Welte <laforge@gnumonks.org>
 *
 * This code is distributed under the terms of GNU GPL v2
 *
 */
#include <getopt.h>
#include <sys/errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <time.h>
#include <netinet/in.h>
#include "libarptc.h"
#include "arptables.h"

static int binary = 0, counters = 0;

static struct option options[] = {
	{ "binary", 0, 0, 'b' },
	{ "counters", 0, 0, 'c' },
	{ "dump", 0, 0, 'd' },
	{ "table", 1, 0, 't' },
	{ 0 }
};

#define PROC_ARP_TABLES_NAMES "/proc/net/arp_tables_names"

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

/* This assumes that mask is contiguous, and byte-bounded. */
static void
print_iface(char letter, const char *iface, const unsigned char *mask,
	    int invert)
{
	unsigned int i;

	if (mask[0] == 0)
		return;

	printf("-%c %s", letter, invert ? "! " : "");

	for (i = 0; i < IFNAMSIZ; i++) {
		if (mask[i] != 0) {
			if (iface[i] != '\0')
				printf("%c", iface[i]);
		} else {
			/* we can access iface[i-1] here, because 
			 * a few lines above we make sure that mask[0] != 0 */
			if (iface[i-1] != '\0')
				printf("+");
			break;
		}
	}

	printf(" ");
}

static void print_uint8 ( const char *opt, u_int8_t value, u_int8_t mask,
 int invert) {
	if (mask == 0)
		return;

	printf("%s %s %02x", opt, invert ? "! " : "", value);

        if ( mask != 0xff )
	    printf ( "/%02x", mask );

	printf(" ");
}

static void print_uint16(const char *opt, u_int16_t value, u_int16_t mask,
 int invert) {
	if (mask == 0)
		return;

	printf("%s %s %04x", opt, invert ? "! " : "", value);

        if ( mask != 0xff )
	    printf ( "/%04x", mask );

	printf(" ");
}

static void print_hwaddr_and_mask ( const char *opt, const char *value,
 const char *mask, int invert ) {
	unsigned int i;
	unsigned int j;
	unsigned int k;

    for ( i = 0; i < IFNAMSIZ; i++ )
	if ( mask[i] != 0 )
	    break;
    if ( i == IFNAMSIZ )
	return;

    for ( j = i+1; j < IFNAMSIZ; j++ )
	if ( mask[j] == 0 )
	    break;
    for ( k = IFNAMSIZ - 1; k > i; k-- )
	if ( mask[k] != 0 )
	    break;

    printf("%s %s", opt, invert ? "! " : "" );

    for ( i = 0; i <= k; i++ )
	printf ( "%s%02x", i > 0 ? ":" : "", value[i] & 0xFF );

    for ( i = 0; i <= k; i++ )
	if ( (mask[i] & 0xFF) != 0xFF )
	    break;

    if ( i != k + 1 )
	for ( i = 0; i <= k; i++ )
	    printf ( "%s%02x", i > 0 ? ":" : "/", mask[i] & 0xFF );

	printf(" ");
}

/* print a given ip including mask if neccessary */
static void print_ip(char *prefix, u_int32_t ip, u_int32_t mask, int invert)
{
	if (!mask && !ip)
		return;

	printf("%s %s%u.%u.%u.%u",
		prefix,
		invert ? "! " : "",
		IP_PARTS(ip));

	if (mask != 0xffffffff) 
		printf("/%u.%u.%u.%u ", IP_PARTS(mask));
	else
		printf(" ");
}

/* We want this to be readable, so only print out neccessary fields.
 * Because that's the kind of world I want to live in.  */
static void print_rule(const struct arpt_entry *e, 
		arptc_handle_t *h, const char *chain, int counters)
{
	struct arpt_entry_target *t;
	const char *target_name;

	/* print counters */
	if (counters)
		printf("[%" FMT_U_INT64_T ":%" FMT_U_INT64_T "] ", e->counters.pcnt, e->counters.bcnt);

	/* print chain name */
	printf("-A %s ", chain);

	/* Print IP part. */
	print_ip("-s", e->arp.src.s_addr,e->arp.smsk.s_addr,
			e->arp.invflags & ARPT_INV_SRCIP);	

	print_ip("-d", e->arp.tgt.s_addr, e->arp.tmsk.s_addr,
			e->arp.invflags & ARPT_INV_TGTIP);

        print_uint8 ( "-a", e->arp.arhln, e->arp.arhln_mask,
 e->arp.invflags & ARPT_INV_ARPHLN );

	print_uint16 ( "-p", e->arp.arpop, e->arp.arpop_mask,
 e->arp.invflags & ARPT_INV_ARPOP );

        print_uint16 ( "-H", e->arp.arhrd, e->arp.arhrd_mask,
 e->arp.invflags & ARPT_INV_ARPHRD );
        print_uint16 ( "-w", e->arp.arpro, e->arp.arpro_mask,
 e->arp.invflags & ARPT_INV_ARPPRO );

        print_hwaddr_and_mask ( "-z", e->arp.src_devaddr.addr, e->arp.src_devaddr.mask,
 e->arp.invflags & ARPT_INV_SRCDEVADDR );

        print_hwaddr_and_mask ( "-y", e->arp.tgt_devaddr.addr, e->arp.tgt_devaddr.mask,
 e->arp.invflags & ARPT_INV_TGTDEVADDR );

	print_iface('i', e->arp.iniface, e->arp.iniface_mask,
		    e->arp.invflags & ARPT_INV_VIA_IN);

	print_iface('o', e->arp.outiface, e->arp.outiface_mask,
		    e->arp.invflags & ARPT_INV_VIA_OUT);

	/* Print target name */	
	target_name = arptc_get_target(e, h);
	if (target_name && (*target_name != '\0'))
		printf("-j %s ", target_name);

	/* Print targinfo part */
	t = arpt_get_target((struct arpt_entry *)e);
	if (t->u.user.name[0]) {
		struct arptables_target *target
			= find_target(t->u.user.name, TRY_LOAD);

		if (!target) {
			fprintf(stderr, "Can't find library for target `%s'\n",
				t->u.user.name);
			exit(1);
		}

		if (target->save)
			target->save(&e->arp, t);
		else {
			/* If the target size is greater than arpt_entry_target
			 * there is something to be saved, we just don't know
			 * how to print it */
			if (t->u.target_size != 
			    sizeof(struct arpt_entry_target)) {
				fprintf(stderr, "Target `%s' is missing "
						"save function\n",
					t->u.user.name);
				exit(1);
			}
		}
	}
	printf("\n");
}

/* Debugging prototype. */
static int for_each_table(int (*func)(const char *tablename))
{
        int ret = 1;
	FILE *procfile = NULL;
	char tablename[ARPT_TABLE_MAXNAMELEN+1];

	procfile = fopen(PROC_ARP_TABLES_NAMES, "r");
	if (!procfile)
		exit_error(OTHER_PROBLEM, 
			   "Can't open %s: %s",
 PROC_ARP_TABLES_NAMES, strerror ( errno ) );


	while (fgets(tablename, sizeof(tablename), procfile)) {
		if (tablename[strlen(tablename) - 1] != '\n')
			exit_error(OTHER_PROBLEM, 
				   "Badly formed tablename `%s'\n",
				   tablename);
		tablename[strlen(tablename) - 1] = '\0';
		ret &= func(tablename);
	}

	return ret;
}
	

static int do_output(const char *tablename)
{
	arptc_handle_t h;
	const char *chain = NULL;

	if (!tablename)
		return for_each_table(&do_output);

	h = arptc_init(tablename);
	if (!h)
 		exit_error(OTHER_PROBLEM, "Can't initialize: %s\n",
			   arptc_strerror(errno));

	if (!binary) {
		time_t now = time(NULL);

		printf("# Generated by arptables-save v%s on %s",
		       ARPTABLES_VERSION, ctime(&now));
		printf("*%s\n", tablename);

		/* Dump out chain names first, 
		 * thereby preventing dependency conflicts */
		for (chain = arptc_first_chain(&h);
		     chain;
		     chain = arptc_next_chain(&h)) {
			
			printf(":%s ", chain);
			if (arptc_builtin(chain, h)) {
				struct arpt_counters count;
				printf("%s ",
				       arptc_get_policy(chain, &count, &h));
				printf("[%" FMT_U_INT64_T ":%" FMT_U_INT64_T "]\n", count.pcnt, count.bcnt);
			} else {
				printf("- [0:0]\n");
			}
		}
				

		for (chain = arptc_first_chain(&h);
		     chain;
		     chain = arptc_next_chain(&h)) {
			const struct arpt_entry *e;

			/* Dump out rules */
			e = arptc_first_rule(chain, &h);
			while(e) {
				print_rule(e, &h, chain, counters);
				e = arptc_next_rule(e, &h);
			}
		}

		now = time(NULL);
		printf("COMMIT\n");
		printf("# Completed on %s", ctime(&now));
	} else {
		/* Binary, huh?  OK. */
		exit_error(OTHER_PROBLEM, "Binary NYI\n");
	}

	return 1;
}

/* Format:
 * :Chain name POLICY packets bytes
 * rule
 */
int main(int argc, char *argv[])
{
	const char *tablename = NULL;
	int c;

	program_name = "arptables-save";
	program_version = ARPTABLES_VERSION;

#ifdef NO_SHARED_LIBS
	init_extensions();
#endif

	while ((c = getopt_long(argc, argv, "bcdt:", options, NULL)) != -1) {
		switch (c) {
		case 'b':
			binary = 1;
			break;

		case 'c':
			counters = 1;
			break;

		case 't':
			/* Select specific table. */
			tablename = optarg;
			break;
		case 'd':
			do_output(tablename);
			exit(0);
		}
	}

	if (optind < argc) {
		fprintf(stderr, "Unknown arguments found on commandline");
		exit(1);
	}

	return !do_output(tablename);
}
