/* Code to take an iptables-style command line and do it. */

/*
 * Author: Paul.Russell@rustcorp.com.au and mneuling@radlogic.com.au
 *
 * (C) 2000-2002 by the netfilter coreteam <coreteam@netfilter.org>:
 * 		    Paul 'Rusty' Russell <rusty@rustcorp.com.au>
 * 		    Marc Boucher <marc+nf@mbsi.ca>
 * 		    James Morris <jmorris@intercode.com.au>
 * 		    Harald Welte <laforge@gnumonks.org>
 * 		    Jozsef Kadlecsik <kadlec@blackhole.kfki.hu>
 * changes Copyright 2003 Red Hat (written by Jay Fenlason)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <getopt.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <ctype.h>
#include <stdarg.h>
#include <limits.h>
#include <unistd.h>
#include "arptables.h"
#include <fcntl.h>
#include <sys/wait.h>

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#ifndef IPT_LIB_DIR
#define IPT_LIB_DIR "/lib/arptables"
#endif

#ifndef PROC_SYS_MODPROBE
#define PROC_SYS_MODPROBE "/proc/sys/kernel/modprobe"
#endif

#define FMT_NUMERIC	0x0001
#define FMT_NOCOUNTS	0x0002
#define FMT_KILOMEGAGIGA 0x0004
/* #define FMT_OPTIONS	0x0008 */
#define FMT_NOTABLE	0x0010
#define FMT_NOTARGET	0x0020
#define FMT_VIA		0x0040
#define FMT_NONEWLINE	0x0080
#define FMT_LINENUMBERS 0x0100

#define FMT_PRINT_RULE (FMT_NOCOUNTS | FMT_VIA \
			| FMT_NUMERIC | FMT_NOTABLE)
#define FMT(tab,notab) ((format) & FMT_NOTABLE ? (notab) : (tab))

#define CMD_NONE		0x0000U
#define CMD_INSERT		0x0001U
#define CMD_DELETE		0x0002U
#define CMD_DELETE_NUM		0x0004U
#define CMD_REPLACE		0x0008U
#define CMD_APPEND		0x0010U
#define CMD_LIST		0x0020U
#define CMD_FLUSH		0x0040U
#define CMD_ZERO		0x0080U
#define CMD_NEW_CHAIN		0x0100U
#define CMD_DELETE_CHAIN	0x0200U
#define CMD_SET_POLICY		0x0400U
#define CMD_CHECK		0x0800U
#define CMD_RENAME_CHAIN	0x1000U
#define NUMBER_OF_CMD	13
static const char cmdflags[] = { 'I', 'D', 'D', 'R', 'A', 'L', 'F', 'Z',
				 'N', 'X', 'P', 'E' };

#define OPTION_OFFSET 256

#define OPT_NONE	0x00000U
#define OPT_NUMERIC	0x00001U
#define OPT_JUMP	0x00002U
#define OPT_VERBOSE	0x00004U
#define OPT_LINENUMBERS 0x00008U
#define OPT_COUNTERS	0x00010U
#define OPT_EXPANDED	0x00020U

#define OPT_SRC_IP	0x00040U
#define OPT_TGT_IP	0x00080U
#define OPT_SRC_DEVA	0x00100U
#define OPT_TGT_DEVA    0x00200U
#define OPT_VIANAMEIN	0x00400U
#define OPT_VIANAMEOUT	0x00800U

#define OPT_ARHLN	0x01000U
#define OPT_ARPOP       0x02000U
#define OPT_ARHRD       0x04000U
#define OPT_ARPRO       0x08000U

#define NUMBER_OF_OPT	16
static const char optflags[NUMBER_OF_OPT]
= { 'n', 'j', 'v', '0', 'c', 'x', 's', 'd', 'z', 'y', 'i', 'o', 'a', 'p', 'H', 'w' };

static struct option original_opts[] = {
	{ "append",        1, 0, 'A' },
	{ "delete",        1, 0, 'D' },
	{ "insert",        1, 0, 'I' },
	{ "replace",       1, 0, 'R' },
	{ "list",          2, 0, 'L' },
	{ "flush",         2, 0, 'F' },
	{ "zero",          2, 0, 'Z' },
	{ "new-chain",     1, 0, 'N' },
	{ "delete-chain",  2, 0, 'X' },
	{ "rename-chain",  2, 0, 'E' },
	{ "policy",        1, 0, 'P' },

	{ "source",        1, 0, 's' },
	{ "destination",   1, 0, 'd' },
	{ "src",           1, 0, 's' }, /* synonym */
	{ "dst",           1, 0, 'd' }, /* synonym */
	{ "target",        1, 0, 'd' }, /* synonym */
	{ "tgt",           1, 0, 'd' }, /* synonym */

	{ "source-hw",     1, 0, 'z' },
	{ "source-mac",    1, 0, 'z' }, /* synonym */
	{ "src-hw",        1, 0, 'z' }, /* synonym */
	{ "src-mac",       1, 0, 'z' }, /* synonym */
	{ "target-hw",     1, 0, 'y' },
	{ "target-mac",    1, 0, 'y' }, /* synonym */
	{ "tgt-mac",       1, 0, 'y' }, /* synonym */
	{ "tgt-hw",        1, 0, 'y' }, /* synonym */
	{ "dst-hw",        1, 0, 'y' }, /* synonym */
	{ "dst-hw",        1, 0, 'y' }, /* synonym */
	{ "in-interface",  1, 0, 'i' },
	{ "out-interface", 1, 0, 'o' },
	{ "arhln",         1, 0, 'a' },
	{ "arpop",         1, 0, 'p' },
	{ "arhrd",         1, 0, 'H' },
	{ "arpro",         1, 0, 'w' },

	{ "verbose",       0, 0, 'v' },
	{ "jump",          1, 0, 'j' },
	{ "table",         1, 0, 't' },
	{ "numeric",       0, 0, 'n' },
	{ "exact",         0, 0, 'x' },
	{ "help",          2, 0, 'h' },
	{ "line-numbers",  0, 0, '0' },
	{ "modprobe",      1, 0, 'M' },
	{ "set-counters",  1, 0, 'c' },
	{ 0 }
};

#if !defined(__OPTIMIZE__) && 0
struct arpt_entry_target *
arpt_get_target(struct arpt_entry *e)
{
	return (void *)e + e->target_offset;
}
#endif

static struct option *opts = original_opts;
static unsigned int global_option_offset = 0;

/* Table of legal combinations of commands and options.  If any of the
 * given commands make an option legal, that option is legal (applies to
 * CMD_LIST and CMD_ZERO only).
 * Key:
 *  +  compulsory
 *  x  illegal
 *     optional
 */

static char commands_v_options[NUMBER_OF_CMD][NUMBER_OF_OPT] =
/* Well, it's better than "Re: Linux vs FreeBSD" */
{
	/*     -n  -j  -v  -0  -c  -x  -s  -d  -z  -y  -i  -o  -a  -p  -H  -w  */
/*INSERT*/    {'x',' ',' ','x',' ','x',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '},
/*DELETE*/    {'x',' ',' ','x',' ','x',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '},
/*DELETE_NUM*/{'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*REPLACE*/   {'x',' ',' ','x',' ','x',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '},
/*APPEND*/    {'x',' ',' ','x',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' ',' '},
/*LIST*/      {' ','x',' ',' ',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*FLUSH*/     {'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*ZERO*/      {'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*NEW_CHAIN*/ {'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*DEL_CHAIN*/ {'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*SET_POLICY*/{'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'},
/*CHECK*/     {'x','x',' ','x',' ','x','+','+','+','+',' ',' ','+',' ',' ',' '},
/*RENAME*/    {'x','x',' ','x',' ','x','x','x','x','x','x','x','x','x','x','x'}
};

static int inverse_for_options[NUMBER_OF_OPT] =
{
/* -n */ 0,
/* -j */ 0,
/* -v */ 0,
/*--line*/ 0,
/* -c */ 0,
/* -x */ 0,
/* -s */ ARPT_INV_SRCIP,
/* -d */ ARPT_INV_TGTIP,
/* -z */ ARPT_INV_SRCDEVADDR,
/* -y */ ARPT_INV_TGTDEVADDR,
/* -i */ ARPT_INV_VIA_IN,
/* -o */ ARPT_INV_VIA_OUT,
/* -a */ ARPT_INV_ARPHLN,
/* -p */ ARPT_INV_ARPOP,
/* -H */ ARPT_INV_ARPHRD,
/* -w */ ARPT_INV_ARPPRO,
};

const char *program_version = ARPTABLES_VERSION;
const char *program_name;

/* Keeping track of external matches and targets: linked lists.  */
struct arptables_target *arptables_targets = NULL;

struct tabentry {
    const char *name;
    u_int16_t value;
};

static struct tabentry opcodes[] =
{
    { "Request",	    0x0001 },
    { "Reply",		    0x0002 },
    { "ReqR",               0x0003 },
    { "Request_Reverse",    0x0003 },
    { "RevR",               0x0004 },
    { "Reply_Reverse",	    0x0004 },
    { "DRRQ",               0x0005 },
    { "DRARP_Request",	    0x0005 },
    { "DRRP",               0x0006 },
    { "DRARP_Reply",	    0x0006 },
    { "DRER",               0x0007 },
    { "DRARP_Error",	    0x0007 },
    { "InARP_Request",	    0x0008 },
    { "ARP_NAK",	    0x0009 },
    { 0,                    0 }
};

static struct tabentry hwtypes[] = {
    { "Ethernet",          0x0001 },
    { 0,                   0 }
};

static struct tabentry protocols[] = {
    { "IPv4",		    0x0800 },
    { 0,                    0 }
};

int
string_to_number(const char *s, unsigned int min, unsigned int max,
		 unsigned int *ret)
{
	long number;
	char *end;

	/* Handle hex, octal, etc. */
	errno = 0;
	number = strtol(s, &end, 0);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (errno != ERANGE && min <= number && number <= max) {
			*ret = number;
			return 0;
		}
	}
	return -1;
}

struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	unsigned int onebyte;
	int i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;

		*q = '\0';
		if (string_to_number(p, 0, 255, &onebyte) == -1)
			return (struct in_addr *) NULL;

		addrp[i] = (unsigned char) onebyte;
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if (string_to_number(p, 0, 255, &onebyte) == -1)
		return (struct in_addr *) NULL;

	addrp[3] = (unsigned char) onebyte;

	return &addr;
}

static struct in_addr *
network_to_addr(const char *name)
{
	struct netent *net;
	static struct in_addr addr;

	if ((net = getnetbyname(name)) != NULL) {
		if (net->n_addrtype != AF_INET)
			return (struct in_addr *) NULL;
		addr.s_addr = htonl((unsigned long) net->n_net);
		return &addr;
	}

	return (struct in_addr *) NULL;
}

static void
inaddrcpy(struct in_addr *dst, struct in_addr *src)
{
	/* memcpy(dst, src, sizeof(struct in_addr)); */
	dst->s_addr = src->s_addr;
}

void
exit_tryhelp(int status)
{
	fprintf(stderr, "Try `%s -h' or '%s --help' for more information.\n",
			program_name, program_name );
	exit(status);
}

void
exit_error(enum exittype status, char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	fprintf(stderr, "%s v%s: ", program_name, program_version);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
	if (status == PARAMETER_PROBLEM)
		exit_tryhelp(status);
	if (status == VERSION_PROBLEM)
		fprintf(stderr,
 "Perhaps iptables or your kernel needs to be upgraded.\n");
	exit(status);
}

void
exit_printhelp(void)
{
	struct arptables_target *t = NULL;

	printf("%s v%s\n\n"
"Usage: %s -[AD] chain rule-specification [options]\n"
"       %s -[RI] chain rulenum rule-specification [options]\n"
"       %s -D chain rulenum [options]\n"
"       %s -[LFZ] [chain] [options]\n"
"       %s -[NX] chain\n"
"       %s -E old-chain-name new-chain-name\n"
"       %s -P chain target [options]\n"
"       %s -h (print this help information)\n\n",
	       program_name, program_version, program_name, program_name,
	       program_name, program_name, program_name, program_name,
	       program_name, program_name);

	printf(
"Commands:\n"
"Either long or short options are allowed.\n"
"  --append  -A chain		Append to chain\n"
"  --delete  -D chain		Delete matching rule from chain\n"
"  --delete  -D chain rulenum\n"
"				Delete rule rulenum (1 = first) from chain\n"
"  --insert  -I chain [rulenum]\n"
"				Insert in chain as rulenum (default 1=first)\n"
"  --replace -R chain rulenum\n"
"				Replace rule rulenum (1 = first) in chain\n"
"  --list    -L [chain]		List the rules in a chain or all chains\n"
"  --flush   -F [chain]		Delete all rules in  chain or all chains\n"
"  --zero    -Z [chain]		Zero counters in chain or all chains\n"
"  --new     -N chain		Create a new user-defined chain\n"
"  --delete-chain\n"
"            -X [chain]		Delete a user-defined chain\n"
"  --policy  -P chain target\n"
"				Change policy on chain to target\n"
"  --rename-chain\n"
"            -E old-chain new-chain\n"
"				Change chain name, (moving any references)\n"

"Options:\n"
"  --source	-s [!] address[/mask]\n"
"				source specification\n"
"  --target     -d [!] address[/mask]\n"
"				destination specification\n"
"  --arhln      -a [!] length[/mask]\n"
"                               hardware address length\n"
"  --source-hw  -z [!] address[/mask]\n"
"                               source hardware address\n"
"  --target-hw  -y [!] address[/mask]\n"
"                               target hardware address\n"
"  --arhln      -a [!] length[/mask]\n"
"                               Hardware address length\n"
"  --arpop      -p [!] operation[/mask]\n"
"                               ARP operation\n"
"  --arhrd      -h [!] hrd[/mask]\n"
"                               ARP hardware address\n"
"  --arpro      -w [!] plen[/mask]\n"
"                               ARP protocol address format\n"
"  --in-interface -i [!] input name[+]\n"
"				network interface name ([+] for wildcard)\n"
"  --jump	-j target\n"
"				target for rule (may load target extension)\n"
"  --numeric	-n		numeric output of addresses and ports\n"
"  --out-interface -o [!] output name[+]\n"
"				network interface name ([+] for wildcard)\n"
"  --table	-t table	table to manipulate (default: `filter')\n"
"  --verbose	-v		verbose mode\n"
"  --line-numbers		print line numbers when listing\n"
"  --exact	-x		expand numbers (display exact values)\n"
"  --modprobe=<command>		try to insert modules using this command\n"
"  --set-counters PKTS BYTES	set the counter during insert/append\n"
"[!] --version	-V		print package version.\n");

	/* Print out any special helps. A user might like to be able
	   to add a --help to the commandline, and see expected
	   results. So we call help for all matches & targets */
	for (t=arptables_targets;t;t=t->next) {
		printf("\n");
		(*t->help)();
	}
	exit(0);
}

static void
generic_opt_check(int command, int options)
{
	int i, j, legal = 0;

	/* Check that commands are valid with options.  Complicated by the
	 * fact that if an option is legal with *any* command given, it is
	 * legal overall (ie. -z and -l).
	 */
	for (i = 0; i < NUMBER_OF_OPT; i++) {
		legal = 0; /* -1 => illegal, 1 => legal, 0 => undecided. */

		for (j = 0; j < NUMBER_OF_CMD; j++) {
			if (!(command & (1<<j)))
				continue;

			if (!(options & (1<<i))) {
				if (commands_v_options[j][i] == '+')
					exit_error(PARAMETER_PROBLEM,
 "You need to supply the `-%c' option for this command\n", optflags[i]);
			} else {
				if (commands_v_options[j][i] != 'x')
					legal = 1;
				else if (legal == 0)
					legal = -1;
			}
		}
		if (legal == -1)
			exit_error(PARAMETER_PROBLEM,
				   "Illegal option `-%c' with this command\n",
				   optflags[i]);
	}
}

static char
opt2char(int option)
{
	const char *ptr;
	for (ptr = optflags; option > 1; option >>= 1, ptr++);

	return *ptr;
}

static char
cmd2char(int option)
{
	const char *ptr;
	for (ptr = cmdflags; option > 1; option >>= 1, ptr++);

	return *ptr;
}

static void
add_command(unsigned int *cmd, const int newcmd, const int othercmds, int invert)
{
	if (invert)
		exit_error(PARAMETER_PROBLEM, "unexpected ! flag");
	if (*cmd & (~othercmds))
		exit_error(PARAMETER_PROBLEM, "Can't use -%c with -%c\n",
			   cmd2char(newcmd), cmd2char(*cmd & (~othercmds)));
	*cmd |= newcmd;
}

int
check_inverse(const char option[], int *invert, int *optind, int argc)
{
	if (option && strcmp(option, "!") == 0) {
		if (*invert)
			exit_error(PARAMETER_PROBLEM,
				   "Multiple `!' flags not allowed");
		*invert = TRUE;
		if (optind) {
			*optind = *optind+1;
			if (argc && *optind > argc)
				exit_error(PARAMETER_PROBLEM,
					   "no argument following `!'");
		}

		return TRUE;
	}
	return FALSE;
}

static void *
fw_calloc(size_t count, size_t size)
{
	void *p;

	if ((p = calloc(count, size)) == NULL) {
		perror("arptables: calloc failed");
		exit(1);
	}
	return p;
}

static void *
fw_malloc(size_t size)
{
	void *p;

	if ((p = malloc(size)) == NULL) {
		perror("arptables: malloc failed");
		exit(1);
	}
	return p;
}

static struct in_addr *
host_to_addr(const char *name, unsigned int *naddr)
{
	struct hostent *host;
	struct in_addr *addr;
	unsigned int i;

	*naddr = 0;
	if ((host = gethostbyname(name)) != NULL) {
		if (host->h_addrtype != AF_INET ||
		    host->h_length != sizeof(struct in_addr))
			return (struct in_addr *) NULL;

		while (host->h_addr_list[*naddr] != (char *) NULL)
			(*naddr)++;
		addr = fw_calloc(*naddr, sizeof(struct in_addr));
		for (i = 0; i < *naddr; i++)
			inaddrcpy(&(addr[i]),
				  (struct in_addr *) host->h_addr_list[i]);
		return addr;
	}

	return (struct in_addr *) NULL;
}

static char *
addr_to_host(const struct in_addr *addr)
{
	struct hostent *host;

	if ((host = gethostbyaddr((char *) addr,
				  sizeof(struct in_addr), AF_INET)) != NULL)
		return (char *) host->h_name;

	return (char *) NULL;
}

/*
 *	All functions starting with "parse" should succeed, otherwise
 *	the program fails.
 *	Most routines return pointers to static data that may change
 *	between calls to the same or other routines with a few exceptions:
 *	"host_to_addr", "parse_hostnetwork", and "parse_hostnetworkmask"
 *	return global static data.
*/

struct in_addr *
parse_hostnetwork(const char *name, unsigned int *naddrs)
{
	struct in_addr *addrp, *addrptmp;

	if ((addrptmp = dotted_to_addr(name)) != NULL ||
	    (addrptmp = network_to_addr(name)) != NULL) {
		addrp = fw_malloc(sizeof(struct in_addr));
		inaddrcpy(addrp, addrptmp);
		*naddrs = 1;
		return addrp;
	}
	if ((addrp = host_to_addr(name, naddrs)) != NULL)
		return addrp;

	exit_error(PARAMETER_PROBLEM, "host/network `%s' not found", name);
	return 0;
}

static struct in_addr *
parse_mask(char *mask)
{
	static struct in_addr maskaddr;
	struct in_addr *addrp;
	unsigned int bits;

	if (mask == NULL) {
		/* no mask at all defaults to 32 bits */
		maskaddr.s_addr = 0xFFFFFFFF;
		return &maskaddr;
	}
	if ((addrp = dotted_to_addr(mask)) != NULL)
		/* dotted_to_addr already returns a network byte order addr */
		return addrp;
	if (string_to_number(mask, 0, 32, &bits) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "invalid mask `%s' specified", mask);
	if (bits != 0) {
		maskaddr.s_addr = htonl(0xFFFFFFFF << (32 - bits));
		return &maskaddr;
	}

	maskaddr.s_addr = 0L;
	return &maskaddr;
}

void
parse_hostnetworkmask(const char *name, struct in_addr **addrpp,
		      struct in_addr *maskp, unsigned int *naddrs)
{
	struct in_addr *addrp;
	char buf[256];
	char *p;
	int i, j, k, n;

	strncpy(buf, name, sizeof(buf) - 1);
	if ((p = strrchr(buf, '/')) != NULL) {
		*p = '\0';
		addrp = parse_mask(p + 1);
	} else
		addrp = parse_mask(NULL);
	inaddrcpy(maskp, addrp);

	/* if a null mask is given, the name is ignored, like in "any/0" */
	if (maskp->s_addr == 0L)
		strcpy(buf, "0.0.0.0");

	addrp = *addrpp = parse_hostnetwork(buf, naddrs);
	n = *naddrs;
	for (i = 0, j = 0; i < n; i++) {
		addrp[j++].s_addr &= maskp->s_addr;
		for (k = 0; k < j - 1; k++) {
			if (addrp[k].s_addr == addrp[j - 1].s_addr) {
				(*naddrs)--;
				j--;
				break;
			}
		}
	}
}

static unsigned digit_to_bits ( char ch ) {
    if ( ch >= '0' && ch <= '9' )
        return ch - '0';
    if ( ch >= 'a' && ch <= 'f' )
	return 10 + ch - 'a';
    if ( ch >= 'A' && ch <= 'F' )
        return 10 + ch - 'A';
    return 0;
}

void parse_uint8 ( const char *name, u_int8_t *value, u_int8_t *mask ) {
	const char *p;
	char ch1, ch2;

	ch1 = name[0];
	ch2 = ch1 ? name[1] : 0;
	if ( isxdigit(ch1) && isxdigit(ch2) ) {
	    *value = (digit_to_bits(ch1)<<4) | digit_to_bits(ch2);
	    p = &name[2];
	} else if ( isxdigit(ch1) ) {
	    *value = digit_to_bits(ch1);
	    p = &name[1];
	} else {
	    *value = 0;
	    p = name;
	}
	if ( *p == '/' ) {
	    p++;
	    ch1 = p[0];
	    ch2 = ch1 ? p[1] : 0;
	    if ( isxdigit(ch1) && isxdigit(ch2) ) {
		*mask = (digit_to_bits(ch1)<<4) | digit_to_bits(ch2);
		p += 2;
	    } else if ( isxdigit(ch1) ) {
		*mask = digit_to_bits(ch1);
		p++;
	    }
	} else
	    *mask = ~0;
        if ( *p )
	    exit_error ( PARAMETER_PROBLEM, "uint8/mask '%s' is invalid",
 name );
}

void parse_uint16 ( const char *name, u_int16_t *value, u_int16_t *mask,
 struct tabentry *table ) {
	const char *p;
	char ch1, ch2, ch3, ch4;
	u_int16_t v;
	u_int16_t m;
	unsigned int i;

	for ( i = 0; table[i].name; i++ ) {
	    if ( ! strcasecmp ( name, table[i].name ) ) {
		*value = table[i].value;
		*mask = 0xFFFF;
		return;
	    }
	}
	if ( strlen ( name ) == 4 ) {
	    for ( i = 0; table[i].name; i++ ) {
		if ( ! strncasecmp ( name, table[i].name, 4 ) ) {
		    *value = table[i].value;
		    *mask = 0xFFFF;
		    return;
		}
	    }
	}

	p = name;
	ch1 = p[0];
	ch2 = ch1 ? p[1] : 0;
	if ( ch1 == '0' && ch1 == 'x' ) {
	    p += 2;
	    ch1 = p[0];
	    ch2 = ch1 ? p[1] : 0;
	}
	ch3 = ch2 ? p[2] : 0;
	ch4 = ch3 ? p[3] : 0;
	if ( isxdigit(ch1) ) {
	    if ( isxdigit(ch2) ) {
	        if ( isxdigit(ch3) ) {
		    if ( isxdigit(ch4) ) {
			v = (digit_to_bits(ch1)<<12) | (digit_to_bits(ch2)<<8) | ( digit_to_bits(ch3)<<4) | digit_to_bits(ch4);
			p = &name[4];
		    } else {
			v = (digit_to_bits(ch1)<<8) | (digit_to_bits(ch2)<<4) | ( digit_to_bits(ch3)<<0);
			p = &name[3];
		    }
		} else {
		    v = (digit_to_bits(ch1)<<4) | (digit_to_bits(ch2)<<0);
		    p = &name[2];
		}
	    } else {
		v = (digit_to_bits(ch1)<<0);
		p = &name[1];
	    }
	} else {
	    v = 0;
	    p = name;
	}
	if ( *p == '/' ) {
	    p++;
	    ch1 = p[0];
	    ch2 = ch1 ? p[1] : 0;
	    if ( ch1 == '0' && ch1 == 'x' ) {
		p += 2;
		ch1 = p[0];
		ch2 = ch1 ? p[1] : 0;
	    }
	    ch3 = ch2 ? p[2] : 0;
	    ch4 = ch3 ? p[3] : 0;
	    if ( isxdigit(ch1) ) {
		if ( isxdigit(ch2) ) {
		    if ( isxdigit(ch3) ) {
			if ( isxdigit(ch4) ) {
			    m = (digit_to_bits(ch1)<<12) | (digit_to_bits(ch2)<<8) | ( digit_to_bits(ch3)<<4) | digit_to_bits(ch4);
			    p += 4;
			} else {
			    m = (digit_to_bits(ch1)<<8) | (digit_to_bits(ch2)<<4) | ( digit_to_bits(ch3)<<0);
			    p += 3;
		        }
		    } else {
		        m = (digit_to_bits(ch1)<<4) | (digit_to_bits(ch2)<<0);
			p += 2;
		    }
		} else {
		    m = (digit_to_bits(ch1)<<0);
		    p++;
		}
	    } else {
		m = 0;
	    }
	} else {
	    m = ~0;
	}
        if ( *p )
	    exit_error ( PARAMETER_PROBLEM, "uint16/mask '%s' is invalid",
 name );
    *value = htons(v);
    *mask = htons(m);
}

void parse_mac_and_mask ( const char *name, struct arpt_devaddr_info *addr ) {
	const char *p;
	int i;
	int j;
	char ch1, ch2;
	char *outp;

        for ( i = 0, p = name, outp = addr->addr; *p; p++) {
	    ch1 = p[0];
	    ch2 = p[1];
	    if ( isxdigit(ch1) && isxdigit(ch2) ) {
		i++;
		*outp++ = (digit_to_bits(ch1)<<4) | digit_to_bits(ch2);
		p += 2;
	    } else if ( isxdigit(ch1) ) {
	        i++;
		*outp++ = digit_to_bits(ch1);
		p++;
	    }
	    if ( *p != ':' ) {
		break;
	    }
	}
	for ( j = i; j++ < ARPT_DEV_ADDR_LEN_MAX; )
	    *outp++ = 0;

	j = 0;
	outp = addr->mask;
	if ( *p == '/' ) {
            for ( p++; *p; p++) {
	        ch1 = p[0];
	        ch2 = p[1];
	        if ( isxdigit(ch1) && isxdigit(ch2) ) {
		    j++;
		    *outp++ = (digit_to_bits(ch1)<<4) | digit_to_bits(ch2);
		    p += 2;
	        } else if ( isxdigit(ch1) ) {
	            j++;
		    *outp++ = digit_to_bits(ch1);
		    p++;
	        }
	        if ( *p != ':' ) {
		    break;
	        }
	    }
	} else {
	    while ( j++ < i )
		*outp++ = ~0;
	}
	while ( j++ < ARPT_DEV_ADDR_LEN_MAX )
	    *outp++ = 0;
        if ( *p )
	    exit_error ( PARAMETER_PROBLEM, "MAC address/mask '%s' is invalid",
 name );
}

static void
parse_interface(const char *arg, char *vianame, unsigned char *mask)
{
	int vialen = strlen(arg);
	unsigned int i;

	memset(mask, 0, IFNAMSIZ);
	memset(vianame, 0, IFNAMSIZ);

	if (vialen + 1 > IFNAMSIZ)
		exit_error(PARAMETER_PROBLEM,
			   "interface name `%s' must be shorter than IFNAMSIZ"
			   " (%i)", arg, IFNAMSIZ-1);

	strcpy(vianame, arg);
	if (vialen == 0)
		memset(mask, 0, IFNAMSIZ);
	else if (vianame[vialen - 1] == '+') {
		memset(mask, 0xFF, vialen - 1);
		memset(mask + vialen - 1, 0, IFNAMSIZ - vialen + 1);
		/* Don't remove `+' here! -HW */
	} else {
		/* Include nul-terminator in match */
		memset(mask, 0xFF, vialen + 1);
		memset(mask + vialen + 1, 0, IFNAMSIZ - vialen - 1);
		for (i = 0; vianame[i]; i++) {
			if (!isalnum(vianame[i]) 
			    && vianame[i] != '_' 
			    && vianame[i] != '.') {
				printf("Warning: wierd character in interface"
				       " `%s' (No aliases, :, ! or *).\n",
				       vianame);
				break;
			}
		}
	}
}

void parse_hwaddr ( const char *name, char *addr, unsigned maxlen ) {
	const char *p;
	unsigned i;
	unsigned j;
	char ch1, ch2;
	char *outp;

        for ( i = 0, p = name, outp = addr; *p; p++) {
	    ch1 = p[0];
	    ch2 = p[1];
	    if ( isxdigit(ch1) && isxdigit(ch2) ) {
		i++;
		*outp++ = (digit_to_bits(ch1)<<4) | digit_to_bits(ch2);
		p += 2;
	    } else if ( isxdigit(ch1) ) {
	        i++;
		*outp++ = digit_to_bits(ch1);
		p++;
	    }
	    if ( *p != ':' ) {
		break;
	    }
	}
	for ( j = i; j++ < maxlen; )
	    *outp++ = 0;
        if ( *p )
	    exit_error ( PARAMETER_PROBLEM, "hardware address '%s' is invalid",
 name );
}

/* Can't be zero. */
static int
parse_rulenumber(const char *rule)
{
	unsigned int rulenum;

	if (string_to_number(rule, 1, INT_MAX, &rulenum) == -1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid rule number `%s'", rule);

	return rulenum;
}

static const char *
parse_target(const char *targetname)
{
	const char *ptr;

	if (strlen(targetname) < 1)
		exit_error(PARAMETER_PROBLEM,
			   "Invalid target name (too short)");

	if (strlen(targetname)+1 > sizeof(arpt_chainlabel))
		exit_error(PARAMETER_PROBLEM,
			   "Invalid target name `%s' (%i chars max)",
			   targetname, sizeof(arpt_chainlabel)-1);

	for (ptr = targetname; *ptr; ptr++)
		if (isspace(*ptr))
			exit_error(PARAMETER_PROBLEM,
				   "Invalid target name `%s'", targetname);
	return targetname;
}

static char *
addr_to_network(const struct in_addr *addr)
{
	struct netent *net;

	if ((net = getnetbyaddr((long) ntohl(addr->s_addr), AF_INET)) != NULL)
		return (char *) net->n_name;

	return (char *) NULL;
}

char *
addr_to_dotted(const struct in_addr *addrp)
{
	static char buf[20];
	const unsigned char *bytep;

	bytep = (const unsigned char *) &(addrp->s_addr);
	sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
	return buf;
}

char *
addr_to_anyname(const struct in_addr *addr)
{
	char *name;

	if ((name = addr_to_host(addr)) != NULL ||
	    (name = addr_to_network(addr)) != NULL)
		return name;

	return addr_to_dotted(addr);
}

char *
mask_to_dotted(const struct in_addr *mask)
{
	int i;
	static char buf[20];
	u_int32_t maskaddr, bits;

	maskaddr = ntohl(mask->s_addr);

	if (maskaddr == 0xFFFFFFFFL)
		/* we don't want to see "/32" */
		return "";

	i = 32;
	bits = 0xFFFFFFFEL;
	while (--i >= 0 && maskaddr != bits)
		bits <<= 1;
	if (i >= 0)
		sprintf(buf, "/%d", i);
	else
		/* mask was not a decent combination of 1's and 0's */
		sprintf(buf, "/%s", addr_to_dotted(mask));

	return buf;
}

static void
set_option(unsigned int *options, unsigned int option, u_int16_t *invflg,
	   int invert)
{
	if (*options & option)
		exit_error(PARAMETER_PROBLEM, "multiple -%c flags not allowed",
			   opt2char(option));
	*options |= option;

	if (invert) {
		unsigned int i;
		for (i = 0; 1 << i != option; i++);

		if (!inverse_for_options[i])
			exit_error(PARAMETER_PROBLEM,
				   "cannot have ! with -%c",
				   opt2char(option));
		*invflg |= inverse_for_options[i];
	}
}

struct arptables_target *
find_target(const char *name, enum arpt_tryload tryload)
{
	struct arptables_target *ptr;

	/* Standard target? */
	if (strcmp(name, "") == 0
	    || strcmp(name, ARPTC_LABEL_ACCEPT) == 0
	    || strcmp(name, ARPTC_LABEL_DROP) == 0
	    || strcmp(name, ARPTC_LABEL_QUEUE) == 0
	    || strcmp(name, ARPTC_LABEL_RETURN) == 0)
		name = "standard";

	for (ptr = arptables_targets; ptr; ptr = ptr->next) {
		if (strcmp(name, ptr->name) == 0)
			break;
	}

#if !defined(NO_SHARED_LIBS)
	if (!ptr && tryload != DONT_LOAD) {
		char path[ 1 + strlen(IPT_LIB_DIR) + strlen("/libarpt_.so")
			 + strlen(name)];
		sprintf(path, IPT_LIB_DIR "/libarpt_%s.so", name);
		if (dlopen(path, RTLD_NOW)) {
			/* Found library.  If it didn't register itself,
			   maybe they specified match as a target. */
			ptr = find_target(name, DONT_LOAD);
			if (!ptr)
				exit_error(PARAMETER_PROBLEM,
					   "Couldn't load target `%s'\n",
					   name);
		} else if (tryload == LOAD_MUST_SUCCEED)
			exit_error(PARAMETER_PROBLEM,
				   "Couldn't load target `%s':%s\n",
				   name, dlerror());
	}
#else
	if (ptr && !ptr->loaded) {
		if (tryload != DONT_LOAD)
			ptr->loaded = 1;
		else
			ptr = NULL;
	}
	if(!ptr && (tryload == LOAD_MUST_SUCCEED)) {
		exit_error(PARAMETER_PROBLEM,
			   "Couldn't find target `%s'\n", name);
	}
#endif

	if (ptr)
		ptr->used = 1;

	return ptr;
}

static struct option *
merge_options(struct option *oldopts, const struct option *newopts,
	      unsigned int *option_offset)
{
	unsigned int num_old, num_new, i;
	struct option *merge;

	for (num_old = 0; oldopts[num_old].name; num_old++);
	for (num_new = 0; newopts[num_new].name; num_new++);

	global_option_offset += OPTION_OFFSET;
	*option_offset = global_option_offset;

	merge = malloc(sizeof(struct option) * (num_new + num_old + 1));
	memcpy(merge, oldopts, num_old * sizeof(struct option));
	for (i = 0; i < num_new; i++) {
		merge[num_old + i] = newopts[i];
		merge[num_old + i].val += *option_offset;
	}
	memset(merge + num_old + num_new, 0, sizeof(struct option));

	return merge;
}

void
register_target(struct arptables_target *me)
{
	if (strcmp(me->version, program_version) != 0) {
		fprintf(stderr, "%s: target `%s' v%s (I'm v%s).\n",
			program_name, me->name, me->version, program_version);
		exit(1);
	}

	if (find_target(me->name, DONT_LOAD)) {
		fprintf(stderr, "%s: target `%s' already registered.\n",
			program_name, me->name);
		exit(1);
	}

	if (me->size != ARPT_ALIGN(me->size)) {
		fprintf(stderr, "%s: target `%s' has invalid size %lu.\n",
			program_name, me->name, (unsigned long)me->size);
		exit(1);
	}

	/* Prepend to list. */
	me->next = arptables_targets;
	arptables_targets = me;
	me->t = NULL;
	me->tflags = 0;
}

static void
print_num(u_int64_t number, unsigned int format)
{
	if (format & FMT_KILOMEGAGIGA) {
		if (number > 99999) {
			number = (number + 500) / 1000;
			if (number > 9999) {
				number = (number + 500) / 1000;
				if (number > 9999) {
					number = (number + 500) / 1000;
					if (number > 9999) {
						number = (number + 500) / 1000;
						printf(FMT("%4" FMT_U_INT64_T "T ","%" FMT_U_INT64_T "T "), number);
					}
					else printf(FMT("%4" FMT_U_INT64_T "G ","%" FMT_U_INT64_T "G "), number);
				}
				else printf(FMT("%4" FMT_U_INT64_T "M ","%" FMT_U_INT64_T "M "), number);
			} else
				printf(FMT("%4" FMT_U_INT64_T "K ","%" FMT_U_INT64_T "K "), number);
		} else
			printf(FMT("%5" FMT_U_INT64_T " ","%" FMT_U_INT64_T " "), number);
	} else
		printf(FMT("%8" FMT_U_INT64_T " ","%" FMT_U_INT64_T " "), number);
}


static void
print_header(unsigned int format, const char *chain, arptc_handle_t *handle)
{
	struct arpt_counters counters;
	const char *pol = arptc_get_policy(chain, &counters, handle);
	printf("Chain %s", chain);
	if (pol) {
		printf(" (policy %s", pol);
		if (!(format & FMT_NOCOUNTS)) {
			fputc(' ', stdout);
			print_num(counters.pcnt, (format|FMT_NOTABLE));
			fputs("packets, ", stdout);
			print_num(counters.bcnt, (format|FMT_NOTABLE));
			fputs("bytes", stdout);
		}
		printf(")\n");
	} else {
		unsigned int refs;
		if (!arptc_get_references(&refs, chain, handle))
			printf(" (ERROR obtaining refs)\n");
		else
			printf(" (%u references)\n", refs);
	}

	if (format & FMT_LINENUMBERS)
		printf(FMT("%-4s ", "%s "), "num");
	if (!(format & FMT_NOCOUNTS)) {
		if (format & FMT_KILOMEGAGIGA) {
			printf(FMT("%5s ","%s "), "pkts");
			printf(FMT("%5s ","%s "), "bytes");
		} else {
			printf(FMT("%8s ","%s "), "pkts");
			printf(FMT("%10s ","%s "), "bytes");
		}
	}
	if (!(format & FMT_NOTARGET))
		printf(FMT("%-9s ","%s "), "target");
	if (format & FMT_VIA) {
		printf(FMT(" %-6s ","%s "), "in");
		printf(FMT("%-6s ","%s "), "out");
	}
	printf(FMT(" %-19s ","%s "), "source-ip");
	printf(FMT(" %-19s "," %s "), "destination-ip");
	printf(FMT(" %-17s ","%s "), "source-hw");
	printf(FMT(" %-17s "," %s "), "destination-hw");
	printf(FMT(" %-5s "," %s "), "hlen" );
	printf(FMT(" %-9s "," %s "), "op" );
	printf(FMT(" %-9s "," %s "), "hrd" );
	printf(FMT(" %-9s "," %s "), "pro" );
	printf("\n");
}

static int hwaddr_is_zero ( const char *hwa ) {
    int i;

    for ( i = 0; i < ARPT_DEV_ADDR_LEN_MAX; i++ )
	if ( hwa[i] )
	    return 0;
    return 1;
}

static char *hwaddr_to_dotted ( const char *hwa, unsigned int u ) {
    static char buf[64];
    char tmp[4];
    unsigned uIndex;

    sprintf ( buf, "%02x", 0xFF & hwa[0] );
    for ( uIndex = 1; uIndex <= u; uIndex++ ) {
	sprintf ( tmp, ":%02x", 0xFF & hwa[uIndex] );
	strcat ( buf, tmp );
    }
    return buf;
}

void print_hwaddr ( const char *hwa, unsigned int u ) {
   printf ( "%s", hwaddr_to_dotted ( hwa, u ) );
}

static void print_uint16 ( unsigned int format, u_int16_t value,
 u_int16_t mask, struct tabentry *names ) {
    unsigned int i;

    if ( format & FMT_NUMERIC ) {
	printf ( FMT("%04x","%x"), value );
	if ( mask != 0xFFFF )
	    printf ( FMT("/%04x ","/%x "), mask );
	else if ( ! ( format & FMT_NOTABLE ) )
	    printf ( "      " );
	return;
    }

    if ( ! mask ) {
	printf ( FMT("%-9s ","%s "), "any" );
	return;
    }
    
    for ( i = 0; names[i].name; i++ )
	if ( names[i].value == value )
	    break;
    if ( ! names[i].name ) {
	print_uint16 ( format | FMT_NUMERIC, value, mask, names );
	return;
    }
    if ( format & FMT_NOTABLE ) {
	printf ( "%s", names[i].name );
	if ( mask != 0xFFFF )
	    printf ( "/%04x", mask );
	return;
    }
    if ( mask == 0xFFFF ) {
	if ( names[i+1].value == value )
	    i++;
	printf ( "%-9.9s ", names[i].name );
	return;
    }
    printf ( "%4.4s/%04x ", names[i].name, mask );
}

/* e is called `fw' here for hysterical raisins */
static void
print_firewall(const struct arpt_entry *fw,
	       const char *targname,
	       unsigned int num,
	       unsigned int format,
	       const arptc_handle_t handle)
{
	struct arptables_target *target = NULL;
	const struct arpt_entry_target *t;
	u_int8_t flags;
	char buf[BUFSIZ];
	unsigned u,v;

	if (!arptc_is_chain(targname, handle))
		target = find_target(targname, TRY_LOAD);
	else
		target = find_target(ARPT_STANDARD_TARGET, LOAD_MUST_SUCCEED);

	t = arpt_get_target((struct arpt_entry *)fw);
	flags = fw->arp.flags;

	if (format & FMT_LINENUMBERS)
		printf(FMT("%-4u ", "%u "), num+1);

	if (!(format & FMT_NOCOUNTS)) {
		print_num(fw->counters.pcnt, format);
		print_num(fw->counters.bcnt, format);
	}

	if (!(format & FMT_NOTARGET))
		printf(FMT("%-9s ", "%s "), targname);

	if (format & FMT_VIA) {
		char iface[IFNAMSIZ+2];

		if (fw->arp.invflags & ARPT_INV_VIA_IN) {
			iface[0] = '!';
			iface[1] = '\0';
		}
		else iface[0] = '\0';

		if (fw->arp.iniface[0] != '\0') {
			strcat(iface, fw->arp.iniface);
		}
		else if (format & FMT_NUMERIC) strcat(iface, "*");
		else strcat(iface, "any");
		printf(FMT(" %-6s ","in %s "), iface);

		if (fw->arp.invflags & ARPT_INV_VIA_OUT) {
			iface[0] = '!';
			iface[1] = '\0';
		}
		else iface[0] = '\0';

		if (fw->arp.outiface[0] != '\0') {
			strcat(iface, fw->arp.outiface);
		}
		else if (format & FMT_NUMERIC) strcat(iface, "*");
		else strcat(iface, "any");
		printf(FMT("%-6s ","out %s "), iface);
	}

	fputc(fw->arp.invflags & ARPT_INV_SRCIP ? '!' : ' ', stdout);
	if (fw->arp.smsk.s_addr == 0L && !(format & FMT_NUMERIC))
		printf(FMT("%-19s ","%s "), "anywhere");
	else {
		if (format & FMT_NUMERIC)
			strcpy ( buf, addr_to_dotted(&(fw->arp.src)));
		else
			strcpy ( buf, addr_to_anyname(&(fw->arp.src)));
		strcat(buf, mask_to_dotted(&(fw->arp.smsk)));
		printf(FMT("%-19s ","%s "), buf);
	}

	fputc(fw->arp.invflags & ARPT_INV_TGTIP ? '!' : ' ', stdout);
	if (fw->arp.tmsk.s_addr == 0L && !(format & FMT_NUMERIC))
		printf(FMT("%-19s ","-> %s"), "anywhere");
	else {
		if (format & FMT_NUMERIC)
			strcpy ( buf, addr_to_dotted(&(fw->arp.tgt)));
		else
			strcpy (buf, addr_to_anyname(&(fw->arp.tgt)));
		strcat(buf, mask_to_dotted(&(fw->arp.tmsk)));
		printf(FMT("%-19s ","-> %s"), buf);
	}

	fputc(fw->arp.invflags & ARPT_INV_SRCDEVADDR ? '!' : ' ', stdout);
	if ( hwaddr_is_zero (fw->arp.src_devaddr.mask) && !(format & FMT_NUMERIC))
		printf(FMT("%-17s ","%s"), "anywhere");
	else {
		for ( u = IFNAMSIZ - 1; u > 0 && fw->arp.src_devaddr.mask[u] == 0; --u )
		    ;
		strcpy ( buf, hwaddr_to_dotted(fw->arp.src_devaddr.addr, u ) );
		for ( v = 0; v <= u; v++ )
		    if ( ( fw->arp.src_devaddr.mask[v] & 0xFF ) != 0xFF )
			break;
		if ( v <= u ) {
		    strcat ( buf, "/" );
		    strcat ( buf, hwaddr_to_dotted ( fw->arp.src_devaddr.mask, u ) );
		}
		printf(FMT("%-17s ","%s"), buf);
	}

	fputc(fw->arp.invflags & ARPT_INV_TGTDEVADDR ? '!' : ' ', stdout);
	if (hwaddr_is_zero (fw->arp.tgt_devaddr.mask) && !(format & FMT_NUMERIC))
		printf(FMT("%-17s ","-> %s"), "anywhere");
	else {
		for ( u = IFNAMSIZ - 1; u > 0 && fw->arp.tgt_devaddr.mask[u] == 0; --u )
		    ;
		strcpy ( buf, hwaddr_to_dotted ( fw->arp.tgt_devaddr.addr, u ) );
		for ( v = 0; v <= u; v++ )
		    if ( ( fw->arp.tgt_devaddr.mask[v] & 0xFF ) != 0xFF )
			break;
		if ( v <= u ) {
		    strcat ( buf, "/" );
		    strcat ( buf, hwaddr_to_dotted ( fw->arp.tgt_devaddr.mask, u ) );
		}
		printf(FMT("%-17s ","-> %s"), buf);
	}

	fputc( fw->arp.invflags & ARPT_INV_ARPHLN ? '!' : ' ', stdout );
	if ( fw->arp.arhln_mask )
	    printf ( "%02x/%02x ", fw->arp.arhln, fw->arp.arhln_mask );
	else
	    printf ( FMT("%-5s ","%s "), "any" );

	fputc ( fw->arp.invflags & ARPT_INV_ARPOP ? '!' : ' ', stdout );
	print_uint16 ( format, ntohs(fw->arp.arpop), ntohs(fw->arp.arpop_mask) , opcodes );

	fputc ( fw->arp.invflags & ARPT_INV_ARPHRD ? '!' : ' ', stdout );
	print_uint16 ( format, ntohs(fw->arp.arhrd), ntohs(fw->arp.arhrd_mask), hwtypes );

	fputc ( fw->arp.invflags & ARPT_INV_ARPPRO ? '!' : ' ', stdout );
	print_uint16 ( format, ntohs(fw->arp.arpro), ntohs(fw->arp.arpro_mask), protocols );

	if (format & FMT_NOTABLE)
		fputs("  ", stdout);

	if (target) {
		if (target->print)
			/* Print the target information. */
			target->print(&fw->arp, t, format & FMT_NUMERIC);
	} else if (t->u.target_size != sizeof(*t))
		printf("[%lu bytes of unknown target data] ",
		       (unsigned long)(t->u.target_size - sizeof(*t)) );

	if (!(format & FMT_NONEWLINE))
		fputc('\n', stdout);
}

static void
print_firewall_line(const struct arpt_entry *fw,
		    const arptc_handle_t h)
{
	struct arpt_entry_target *t;

	t = arpt_get_target((struct arpt_entry *)fw);
	print_firewall(fw, t->u.user.name, 0, FMT_PRINT_RULE, h);
}

static int
append_entry(const arpt_chainlabel chain,
	     struct arpt_entry *fw,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose,
	     arptc_handle_t *handle)
{
	unsigned int i, j;
	int ret = 1;

	for (i = 0; i < nsaddrs; i++) {
		fw->arp.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			fw->arp.tgt.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall_line(fw, *handle);
			ret &= arptc_append_entry(chain, fw, handle);
		}
	}

	return ret;
}

static int
replace_entry(const arpt_chainlabel chain,
	      struct arpt_entry *fw,
	      unsigned int rulenum,
	      const struct in_addr *saddr,
	      const struct in_addr *daddr,
	      int verbose,
	      arptc_handle_t *handle)
{
	fw->arp.src.s_addr = saddr->s_addr;
	fw->arp.tgt.s_addr = daddr->s_addr;

	if (verbose)
		print_firewall_line(fw, *handle);
	return arptc_replace_entry(chain, fw, rulenum, handle);
}

static int
insert_entry(const arpt_chainlabel chain,
	     struct arpt_entry *fw,
	     unsigned int rulenum,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose,
	     arptc_handle_t *handle)
{
	unsigned int i, j;
	int ret = 1;

	for (i = 0; i < nsaddrs; i++) {
		fw->arp.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			fw->arp.tgt.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall_line(fw, *handle);
			ret &= arptc_insert_entry(chain, fw, rulenum, handle);
		}
	}

	return ret;
}

static unsigned char *
make_delete_mask(struct arpt_entry *fw)
{
	/* Establish mask for comparison */
	unsigned int size;
	unsigned char *mask, *mptr;

	size = sizeof(struct arpt_entry);

	mask = fw_calloc(1, size
			 + ARPT_ALIGN(sizeof(struct arpt_entry_target))
			 + arptables_targets->size);

	memset(mask, 0xFF, sizeof(struct arpt_entry));
	mptr = mask + sizeof(struct arpt_entry);

	memset(mptr, 0xFF,
	       ARPT_ALIGN(sizeof(struct arpt_entry_target))
	       + arptables_targets->userspacesize);

	return mask;
}

static int
delete_entry(const arpt_chainlabel chain,
	     struct arpt_entry *fw,
	     unsigned int nsaddrs,
	     const struct in_addr saddrs[],
	     unsigned int ndaddrs,
	     const struct in_addr daddrs[],
	     int verbose,
	     arptc_handle_t *handle)
{
	unsigned int i, j;
	int ret = 1;
	unsigned char *mask;

	mask = make_delete_mask(fw);
	for (i = 0; i < nsaddrs; i++) {
		fw->arp.src.s_addr = saddrs[i].s_addr;
		for (j = 0; j < ndaddrs; j++) {
			fw->arp.tgt.s_addr = daddrs[j].s_addr;
			if (verbose)
				print_firewall_line(fw, *handle);
			ret &= arptc_delete_entry(chain, fw, mask, handle);
		}
	}
	return ret;
}

int
for_each_chain(int (*fn)(const arpt_chainlabel, int, arptc_handle_t *),
	       int verbose, int builtinstoo, arptc_handle_t *handle)
{
        int ret = 1;
	const char *chain;
	char *chains;
	unsigned int i, chaincount = 0;

	chain = arptc_first_chain(handle);
	while (chain) {
		chaincount++;
		chain = arptc_next_chain(handle);
        }

	chains = fw_malloc(sizeof(arpt_chainlabel) * chaincount);
	i = 0;
	chain = arptc_first_chain(handle);
	while (chain) {
		strcpy(chains + i*sizeof(arpt_chainlabel), chain);
		i++;
		chain = arptc_next_chain(handle);
        }

	for (i = 0; i < chaincount; i++) {
		if (!builtinstoo
		    && arptc_builtin(chains + i*sizeof(arpt_chainlabel),
				    *handle))
			continue;
	        ret &= fn(chains + i*sizeof(arpt_chainlabel), verbose, handle);
	}

	free(chains);
        return ret;
}

int
flush_entries(const arpt_chainlabel chain, int verbose,
	      arptc_handle_t *handle)
{
	if (!chain)
		return for_each_chain(flush_entries, verbose, 1, handle);

	if (verbose)
		fprintf(stdout, "Flushing chain `%s'\n", chain);
	return arptc_flush_entries(chain, handle);
}

static int
zero_entries(const arpt_chainlabel chain, int verbose,
	     arptc_handle_t *handle)
{
	if (!chain)
		return for_each_chain(zero_entries, verbose, 1, handle);

	if (verbose)
		fprintf(stdout, "Zeroing chain `%s'\n", chain);
	return arptc_zero_entries(chain, handle);
}

int
delete_chain(const arpt_chainlabel chain, int verbose,
	     arptc_handle_t *handle)
{
	if (!chain)
		return for_each_chain(delete_chain, verbose, 0, handle);

	if (verbose)
	        fprintf(stdout, "Deleting chain `%s'\n", chain);
	return arptc_delete_chain(chain, handle);
}

static int
list_entries(const arpt_chainlabel chain, int verbose, int numeric,
	     int expanded, int linenumbers, arptc_handle_t *handle)
{
	int found = 0;
	unsigned int format;
	const char *this;

	format = 0;
	if (!verbose)
		format |= FMT_NOCOUNTS;
	else
		format |= FMT_VIA;

	if (numeric)
		format |= FMT_NUMERIC;

	if (!expanded)
		format |= FMT_KILOMEGAGIGA;

	if (linenumbers)
		format |= FMT_LINENUMBERS;

	for (this = arptc_first_chain(handle);
	     this;
	     this = arptc_next_chain(handle)) {
		const struct arpt_entry *i;
		unsigned int num;

		if (chain && strcmp(chain, this) != 0)
			continue;

		if (found) printf("\n");

		print_header(format, this, handle);
		i = arptc_first_rule(this, handle);

		num = 0;
		while (i) {
			print_firewall(i,
				       arptc_get_target(i, handle),
				       num++,
				       format,
				       *handle);
			i = arptc_next_rule(i, handle);
		}
		found = 1;
	}

	errno = ENOENT;
	return found;
}

static char *get_modprobe(void)
{
	int procfile;
	char *ret;

	procfile = open(PROC_SYS_MODPROBE, O_RDONLY);
	if (procfile < 0)
		return NULL;

	ret = malloc(1024);
	if (ret) {
		switch (read(procfile, ret, 1024)) {
		case -1: goto fail;
		case 1024: goto fail; /* Partial read.  Wierd */
		}
		if (ret[strlen(ret)-1]=='\n') 
			ret[strlen(ret)-1]=0;
		close(procfile);
		return ret;
	}
 fail:
	free(ret);
	close(procfile);
	return NULL;
}

int arptables_insmod(const char *modname, const char *modprobe)
{
	char *buf = NULL;
	char *argv[3];

	/* If they don't explicitly set it, read out of kernel */
	if (!modprobe) {
		buf = get_modprobe();
		if (!buf)
			return -1;
		modprobe = buf;
	}

	switch (fork()) {
	case 0:
		argv[0] = (char *)modprobe;
		argv[1] = (char *)modname;
		argv[2] = NULL;
		execv(argv[0], argv);

		/* not usually reached */
		exit(0);
	case -1:
		return -1;

	default: /* parent */
		wait(NULL);
	}

	free(buf);
	return 0;
}

static struct arpt_entry *
generate_entry(const struct arpt_entry *fw, struct arpt_entry_target *target) {
	unsigned int size;
	struct arpt_entry *e;

	size = sizeof(struct arpt_entry);

	e = fw_malloc(size + target->u.target_size);
	*e = *fw;
	e->target_offset = size;
	e->next_offset = size + target->u.target_size;

	size = 0;
	memcpy(e->elems + size, target, target->u.target_size);

	return e;
}

int do_commandarp(int argc, char *argv[], char **table, arptc_handle_t *handle)
{
	struct arpt_entry fw, *e = NULL;
	int invert = 0;
	unsigned int nsaddrs = 0, ndaddrs = 0;
	struct in_addr *saddrs = NULL, *daddrs = NULL;

	int c, verbose = 0;
	const char *chain = NULL;
	const char *shostnetworkmask = NULL, *dhostnetworkmask = NULL;
	const char *smac = NULL, *tmac = NULL;
	const char *arhln = NULL;
	const char *arpop = NULL;
	const char *arhrd = NULL;
	const char *arpro = NULL;
	const char *policy = NULL, *newname = NULL;
	unsigned int rulenum = 0, options = 0, command = 0;
	const char *pcnt = NULL, *bcnt = NULL;
	int ret = 1;
	struct arptables_target *target = NULL;
	struct arptables_target *t;
	const char *jumpto = "";
	const char *modprobe = NULL;
	int status;

	memset(&fw, 0, sizeof(fw));

	opts = original_opts;
	global_option_offset = 0;

	/* re-set optind to 0 in case do_command gets called
	 * a second time */
	optind = 0;

	for (t = arptables_targets; t; t = t->next) {
		t->tflags = 0;
		t->used = 0;
	}

	/* Suppress error messages: we may add new options if we
           demand-load a protocol. */
	opterr = 0;

	while ((c = getopt_long(argc, argv,
	   "-A:D:R:I:L::M:F::Z::N:X::E:P:Vh::o:p:s:d:j:i:fbvnt:m:xc:z:y:a:H:w:",
					   opts, NULL)) != -1) {
		switch (c) {
			/*
			 * Command selection
			 */
		case 'A':
			add_command(&command, CMD_APPEND, CMD_NONE,
				    invert);
			chain = optarg;
			break;

		case 'D':
			add_command(&command, CMD_DELETE, CMD_NONE,
				    invert);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!') {
				rulenum = parse_rulenumber(argv[optind++]);
				command = CMD_DELETE_NUM;
			}
			break;

		case 'R':
			add_command(&command, CMD_REPLACE, CMD_NONE,
				    invert);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				rulenum = parse_rulenumber(argv[optind++]);
			else
				exit_error(PARAMETER_PROBLEM,
					   "-%c requires a rule number",
					   cmd2char(CMD_REPLACE));
			break;

		case 'I':
			add_command(&command, CMD_INSERT, CMD_NONE,
				    invert);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				rulenum = parse_rulenumber(argv[optind++]);
			else rulenum = 1;
			break;

		case 'L':
			add_command(&command, CMD_LIST, CMD_ZERO,
				    invert);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				chain = argv[optind++];
			break;

		case 'F':
			add_command(&command, CMD_FLUSH, CMD_NONE,
				    invert);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				chain = argv[optind++];
			break;

		case 'Z':
			add_command(&command, CMD_ZERO, CMD_LIST,
				    invert);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				&& argv[optind][0] != '!')
				chain = argv[optind++];
			break;

		case 'N':
			if (optarg && *optarg == '-')
				exit_error(PARAMETER_PROBLEM,
					   "chain name not allowed to start "
					   "with `-'\n");
			if (find_target(optarg, TRY_LOAD))
				exit_error(PARAMETER_PROBLEM,
					   "chain name may not clash "
					   "with target name\n");
			add_command(&command, CMD_NEW_CHAIN, CMD_NONE,
				    invert);
			chain = optarg;
			break;

		case 'X':
			add_command(&command, CMD_DELETE_CHAIN, CMD_NONE,
				    invert);
			if (optarg) chain = optarg;
			else if (optind < argc && argv[optind][0] != '-'
				 && argv[optind][0] != '!')
				chain = argv[optind++];
			break;

		case 'E':
			add_command(&command, CMD_RENAME_CHAIN, CMD_NONE,
				    invert);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				newname = argv[optind++];
			else
				exit_error(PARAMETER_PROBLEM,
				           "-%c requires old-chain-name and "
					   "new-chain-name",
					    cmd2char(CMD_RENAME_CHAIN));
			break;

		case 'P':
			add_command(&command, CMD_SET_POLICY, CMD_NONE,
				    invert);
			chain = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				policy = argv[optind++];
			else
				exit_error(PARAMETER_PROBLEM,
					   "-%c requires a chain and a policy",
					   cmd2char(CMD_SET_POLICY));
			break;

		case 'h':
			if (!optarg)
				optarg = argv[optind];
			exit_printhelp();

		case 'a':
			check_inverse ( optarg, &invert, &optind, argc );
			set_option ( &options, OPT_ARHLN, &fw.arp.invflags, invert );
			arhln = argv[optind - 1];
			break;

		case 'p':
			check_inverse ( optarg, &invert, &optind, argc );
			set_option ( &options, OPT_ARPOP, &fw.arp.invflags, invert );
			arpop = argv[optind - 1];
			break;

		case 'H':
			check_inverse ( optarg, &invert, &optind, argc );
			set_option ( &options, OPT_ARHRD, &fw.arp.invflags, invert );
			arhrd = argv[optind - 1];
			break;

		case 'w':
			check_inverse ( optarg, &invert, &optind, argc );
			set_option ( &options, OPT_ARPRO, &fw.arp.invflags, invert );
			arpro = argv[optind - 1];
			break;

                case 'z':
		        check_inverse ( optarg, &invert, &optind, argc );
			set_option ( &options, OPT_SRC_DEVA, &fw.arp.invflags,
 invert );
                        smac = argv[optind-1];
			break;

		case 'y':
		        check_inverse ( optarg, &invert, &optind, argc );
			set_option ( &options, OPT_TGT_DEVA, &fw.arp.invflags,
 invert );
                        tmac = argv[optind-1];
			break;

		case 's':
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, OPT_SRC_IP, &fw.arp.invflags,
				   invert);
			shostnetworkmask = argv[optind-1];
			/* fw.nfcache |= NFC_IP_SRC; */
			break;

		case 'd':
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, OPT_TGT_IP, &fw.arp.invflags,
				   invert);
			dhostnetworkmask = argv[optind-1];
			/* fw.nfcache |= NFC_IP_DST; */
			break;

		case 'j':
			set_option(&options, OPT_JUMP, &fw.arp.invflags,
				   invert);
			jumpto = parse_target(optarg);
			/* TRY_LOAD (may be chain name) */
			target = find_target(jumpto, TRY_LOAD);

			if (target) {
				size_t size;

				size = ARPT_ALIGN(sizeof(struct arpt_entry_target))
					+ target->size;

				target->t = fw_calloc(1, size);
				target->t->u.target_size = size;
				strcpy(target->t->u.user.name, jumpto);
				target->init(target->t, NULL /* &fw.nfcache */);
				opts = merge_options(opts, target->extra_opts, &target->option_offset);
			}
			break;


		case 'i':
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, OPT_VIANAMEIN, &fw.arp.invflags,
				   invert);
			parse_interface(argv[optind-1],
					fw.arp.iniface,
					fw.arp.iniface_mask);
			/* fw.nfcache |= NFC_IP_IF_IN; */
			break;

		case 'o':
			check_inverse(optarg, &invert, &optind, argc);
			set_option(&options, OPT_VIANAMEOUT, &fw.arp.invflags,
				   invert);
			parse_interface(argv[optind-1],
					fw.arp.outiface,
					fw.arp.outiface_mask);
			/* fw.nfcache |= NFC_IP_IF_OUT; */
			break;

#if 0
		case 'f':
			set_option(&options, OPT_FRAGMENT, &fw.arp.invflags,
				   invert);
			fw.arp.flags |= ARPT_F_FRAG;
			fw.nfcache |= NFC_IP_FRAG;
			break;
#endif
		case 'v':
			if (!verbose)
				set_option(&options, OPT_VERBOSE,
					   &fw.arp.invflags, invert);
			verbose++;
			break;

		case 'n':
			set_option(&options, OPT_NUMERIC, &fw.arp.invflags,
				   invert);
			break;

		case 't':
			if (invert)
				exit_error(PARAMETER_PROBLEM,
					   "unexpected ! flag before --table");
			*table = argv[optind-1];
			break;

		case 'x':
			set_option(&options, OPT_EXPANDED, &fw.arp.invflags,
				   invert);
			break;

		case 'V':
			if (invert)
				printf("Not %s ;-)\n", program_version);
			else
				printf("%s v%s\n",
				       program_name, program_version);
			exit(0);

		case '0':
			set_option(&options, OPT_LINENUMBERS, &fw.arp.invflags,
				   invert);
			break;

		case 'M':
			modprobe = optarg;
			break;

		case 'c':

			set_option(&options, OPT_COUNTERS, &fw.arp.invflags,
				   invert);
			pcnt = optarg;
			if (optind < argc && argv[optind][0] != '-'
			    && argv[optind][0] != '!')
				bcnt = argv[optind++];
			else
				exit_error(PARAMETER_PROBLEM,
					"-%c requires packet and byte counter",
					opt2char(OPT_COUNTERS));

			if (sscanf(pcnt, "%" FMT_U_INT64_T, &fw.counters.pcnt) != 1)
				exit_error(PARAMETER_PROBLEM,
					"-%c packet counter not numeric",
					opt2char(OPT_COUNTERS));

			if (sscanf(bcnt, "%" FMT_U_INT64_T, &fw.counters.bcnt) != 1)
				exit_error(PARAMETER_PROBLEM,
					"-%c byte counter not numeric",
					opt2char(OPT_COUNTERS));
			
			break;


		case 1: /* non option */
			if (optarg[0] == '!' && optarg[1] == '\0') {
				if (invert)
					exit_error(PARAMETER_PROBLEM,
						   "multiple consecutive ! not"
						   " allowed");
				invert = TRUE;
				optarg[0] = '\0';
				continue;
			}
			printf("Bad argument `%s'\n", optarg);
			exit_tryhelp(2);

		default:
			if ( target ) {
			    status = (*target->parse) (
 c - target->option_offset, argv, invert, &(target->tflags), &fw, &(target->t) );
			    invert = FALSE;
			} else
			    status = 1;
			if ( status )
			    exit_error(PARAMETER_PROBLEM, "Unknown arg `%s'",
 argv[optind-1]);
		}
		invert = FALSE;
	}

	if (target)
		(*target->final_check)(target->tflags);

	/* Fix me: must put inverse options checking here --MN */

	if (optind < argc)
		exit_error(PARAMETER_PROBLEM,
			   "unknown arguments found on commandline");
	if (!command)
		exit_error(PARAMETER_PROBLEM, "no command specified");
	if (invert)
		exit_error(PARAMETER_PROBLEM,
			   "nothing appropriate following !");

	if (command & (CMD_REPLACE | CMD_INSERT | CMD_DELETE | CMD_APPEND)) {
		if (!(options & OPT_TGT_IP))
			dhostnetworkmask = "0.0.0.0/0";
		if (!(options & OPT_SRC_IP))
			shostnetworkmask = "0.0.0.0/0";
		if (!(options & OPT_TGT_DEVA))
			tmac = "0/0";
		if (!(options & OPT_SRC_DEVA))
			smac = "0/0";
	        if ( ! ( options & OPT_ARHLN ) )
			arhln = "0/0";
	        if ( ! ( options & OPT_ARPOP ) )
			arpop = "0/0";
	}

	if (shostnetworkmask)
		parse_hostnetworkmask(shostnetworkmask, &saddrs,
				      &(fw.arp.smsk), &nsaddrs);

	if (dhostnetworkmask)
		parse_hostnetworkmask(dhostnetworkmask, &daddrs,
				      &(fw.arp.tmsk), &ndaddrs);

        if ( smac )
	    parse_mac_and_mask ( smac, &fw.arp.src_devaddr );

        if ( tmac )
	    parse_mac_and_mask ( tmac, &fw.arp.tgt_devaddr );

	if ( arhln )
	    parse_uint8 ( arhln, &fw.arp.arhln, &fw.arp.arhln_mask );

	if ( arpop )
	    parse_uint16 ( arpop, &fw.arp.arpop, &fw.arp.arpop_mask, opcodes );

	if ( arhrd )
	    parse_uint16 ( arhrd, &fw.arp.arhrd, &fw.arp.arhrd_mask, hwtypes );

	if ( arpro )
	    parse_uint16 ( arpro, &fw.arp.arpro, &fw.arp.arpro_mask, protocols );

	if ((nsaddrs > 1 || ndaddrs > 1) &&
	    (fw.arp.invflags & (ARPT_INV_SRCIP | ARPT_INV_TGTIP)))
		exit_error(PARAMETER_PROBLEM, "! not allowed with multiple"
			   " source or destination IP addresses");

	if (command == CMD_REPLACE && (nsaddrs != 1 || ndaddrs != 1))
		exit_error(PARAMETER_PROBLEM, "Replacement rule does not "
			   "specify a unique address");

	generic_opt_check(command, options);

	if (chain && strlen(chain) > ARPT_FUNCTION_MAXNAMELEN)
		exit_error(PARAMETER_PROBLEM,
			   "chain name `%s' too long (must be under %i chars)",
			   chain, ARPT_FUNCTION_MAXNAMELEN);

	/* only allocate handle if we weren't called with a handle */
	if (!*handle)
		*handle = arptc_init(*table);

	if (!*handle) {
		/* try to insmod the module if arptc_init failed */
		arptables_insmod("arp_tables", modprobe);
		*handle = arptc_init(*table);
	}

	if (!*handle)
		exit_error(VERSION_PROBLEM,
			   "can't initialize arptables table `%s': %s",
			   *table, arptc_strerror(errno));

	if (command == CMD_APPEND
	    || command == CMD_DELETE
	    || command == CMD_INSERT
	    || command == CMD_REPLACE) {
		if ( strcmp(chain, "INPUT") == 0) {
			/* -o not valid with incoming packets. */
			if (options & OPT_VIANAMEOUT)
				exit_error(PARAMETER_PROBLEM,
					   "Can't use -%c with %s\n",
					   opt2char(OPT_VIANAMEOUT),
					   chain);
		}

		if ( strcmp(chain, "OUTPUT") == 0) {
			/* -i not valid with outgoing packets */
			if (options & OPT_VIANAMEIN)
				exit_error(PARAMETER_PROBLEM,
					   "Can't use -%c with %s\n",
					   opt2char(OPT_VIANAMEIN),
					   chain);
		}

		if (target && arptc_is_chain(jumpto, *handle)) {
			printf("Warning: using chain %s, not extension\n",
			       jumpto);

			target = NULL;
		}

		/* If they didn't specify a target, or it's a chain
		   name, use standard. */
		if (!target
		    && (strlen(jumpto) == 0
			|| arptc_is_chain(jumpto, *handle))) {
			size_t size;

			target = find_target(ARPT_STANDARD_TARGET,
					     LOAD_MUST_SUCCEED);

			size = sizeof(struct arpt_entry_target)
				+ target->size;
			target->t = fw_calloc(1, size);
			target->t->u.target_size = size;
			strcpy(target->t->u.user.name, jumpto);
			(*target->init)(target->t, NULL /* &fw.nfcache */);
		}

		if (!target) {
			/* it is no chain, and we can't load a plugin.
			 * We cannot know if the plugin is corrupt, non
			 * existant OR if the user just misspelled a
			 * chain. */
			find_target(jumpto, LOAD_MUST_SUCCEED);
		} else {
			e = generate_entry(&fw, target->t);
		}
	}

	switch (command) {
	case CMD_APPEND:
		ret = append_entry(chain, e,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   handle);
		break;
	case CMD_DELETE:
		ret = delete_entry(chain, e,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   handle);
		break;
	case CMD_DELETE_NUM:
		ret = arptc_delete_num_entry(chain, rulenum - 1, handle);
		break;
	case CMD_REPLACE:
		ret = replace_entry(chain, e, rulenum - 1,
				    saddrs, daddrs, options&OPT_VERBOSE,
				    handle);
		break;
	case CMD_INSERT:
		ret = insert_entry(chain, e, rulenum - 1,
				   nsaddrs, saddrs, ndaddrs, daddrs,
				   options&OPT_VERBOSE,
				   handle);
		break;
	case CMD_LIST:
		ret = list_entries(chain,
				   options&OPT_VERBOSE,
				   options&OPT_NUMERIC,
				   options&OPT_EXPANDED,
				   options&OPT_LINENUMBERS,
				   handle);
		break;
	case CMD_FLUSH:
		ret = flush_entries(chain, options&OPT_VERBOSE, handle);
		break;
	case CMD_ZERO:
		ret = zero_entries(chain, options&OPT_VERBOSE, handle);
		break;
	case CMD_LIST|CMD_ZERO:
		ret = list_entries(chain,
				   options&OPT_VERBOSE,
				   options&OPT_NUMERIC,
				   options&OPT_EXPANDED,
				   options&OPT_LINENUMBERS,
				   handle);
		if (ret)
			ret = zero_entries(chain,
					   options&OPT_VERBOSE, handle);
		break;
	case CMD_NEW_CHAIN:
		ret = arptc_create_chain(chain, handle);
		break;
	case CMD_DELETE_CHAIN:
		ret = delete_chain(chain, options&OPT_VERBOSE, handle);
		break;
	case CMD_RENAME_CHAIN:
		ret = arptc_rename_chain(chain, newname,	handle);
		break;
	case CMD_SET_POLICY:
		ret = arptc_set_policy(chain, policy, NULL, handle);
		break;
	default:
		/* We should never reach this... */
		exit_tryhelp(2);
	}
	return ret;
}
