#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <getopt.h>
#include "arptables.h"
#include "arpt_mangle.h"

static void
help(void)
{
	printf(
"mangle target v%s options:\n"
"--mangle-ip-s IP address\n"
"--mangle-ip-d IP address\n"
"--mangle-hw-s hardware (MAC) address\n"
"--mangle-hw-d hardware (MAC) address\n"
"--mangle-target target (DROP, CONTINUE or ACCEPT -- default is ACCEPT)\n",
	ARPTABLES_VERSION);
}

#define MANGLE_IPS    '1'
#define MANGLE_IPT    '2'
#define MANGLE_DEVS   '3'
#define MANGLE_DEVT   '4'
#define MANGLE_TARGET '5'
static struct option opts[] = {
	{ "mangle-ip-s"   , required_argument, 0, MANGLE_IPS    },
	{ "mangle-ip-d"   , required_argument, 0, MANGLE_IPT    },
	{ "mangle-hw-s"   , required_argument, 0, MANGLE_DEVS   },
	{ "mangle-hw-d"   , required_argument, 0, MANGLE_DEVT   },
	{ "mangle-target" , required_argument, 0, MANGLE_TARGET },
	{0}
};

static void
init(struct arpt_entry_target *t, unsigned int *u __attribute__((__unused__)) )
{
	struct arpt_mangle *mangle = (struct arpt_mangle *) t->data;

	mangle->target = NF_ACCEPT;
}

static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct arpt_entry *e,
      struct arpt_entry_target **t)
{
	struct arpt_mangle *mangle = (struct arpt_mangle *)(*t)->data;
	struct in_addr *ipaddr;

	switch (c) {
	case MANGLE_IPS:
		{
			unsigned int nr;
			ipaddr = parse_hostnetwork(argv[optind-1], &nr);
		}
		mangle->u_s.src_ip.s_addr = ipaddr->s_addr;
		free(ipaddr);
		mangle->flags |= ARPT_MANGLE_SIP;
		break;
	case MANGLE_IPT:
		{
			unsigned int nr;
			ipaddr = parse_hostnetwork(argv[optind-1], &nr);
		}
		mangle->u_t.tgt_ip.s_addr = ipaddr->s_addr;
		free(ipaddr);
		mangle->flags |= ARPT_MANGLE_TIP;
		break;
	case MANGLE_DEVS:
		if (e->arp.arhln_mask == 0)
			exit_error(PARAMETER_PROBLEM, "no hln defined");
		if (e->arp.invflags & ARPT_INV_ARPHLN)
			exit_error(PARAMETER_PROBLEM,
				   "! hln not allowed for --mangle-mac-s");
		if (e->arp.arhln != 6)
			exit_error(PARAMETER_PROBLEM, "only hln=6 supported");
		parse_hwaddr ( argv[optind - 1], mangle->src_devaddr, e->arp.arhln );
		mangle->flags |= ARPT_MANGLE_SDEV;
		break;
	case MANGLE_DEVT:
		if (e->arp.arhln_mask == 0)
			exit_error(PARAMETER_PROBLEM, "no hln defined");
		if (e->arp.invflags & ARPT_INV_ARPHLN)
			exit_error(PARAMETER_PROBLEM,
				   "! hln not allowed for --mangle-mac-d");
		if (e->arp.arhln != 6)
			exit_error(PARAMETER_PROBLEM, "only hln=6 supported");
		parse_hwaddr ( argv[optind - 1], mangle->tgt_devaddr, e->arp.arhln );
		mangle->flags |= ARPT_MANGLE_TDEV;
		break;
	case MANGLE_TARGET:
		if (!strcmp(argv[optind-1], "DROP"))
			mangle->target = NF_DROP;
		else if (!strcmp(argv[optind-1], "ACCEPT"))
			mangle->target = NF_ACCEPT;
		else if (!strcmp(argv[optind-1], "CONTINUE"))
			mangle->target = ARPT_CONTINUE;
		else
			exit_error(PARAMETER_PROBLEM, "bad target for "
						      "--mangle-target");
	}
	return 0;
}

static void final_check(unsigned int flags)
{
}

static void print(const struct arpt_arp *ip,
   const struct arpt_entry_target *target, int numeric)
{
	struct arpt_mangle *m = (struct arpt_mangle *)(target->data);

	if (m->flags & ARPT_MANGLE_SIP) {
		printf("--mangle-ip-s %s ", numeric ? addr_to_dotted(&(m->u_s.src_ip)) : addr_to_anyname(&(m->u_s.src_ip)) );
	}
	if (m->flags & ARPT_MANGLE_SDEV) {
		printf("--mangle-mac-s ");
		print_hwaddr(m->src_devaddr, 6);
		printf(" ");
	}
	if (m->flags & ARPT_MANGLE_TIP) {
		printf("--mangle-ip-d %s ", numeric ? addr_to_dotted(&(m->u_t.tgt_ip)) : addr_to_anyname(&(m->u_t.tgt_ip)) );
	}
	if (m->flags & ARPT_MANGLE_TDEV) {
		printf("--mangle-mac-d ");
		print_hwaddr(m->tgt_devaddr, 6);
		printf(" ");
	}
	if (m->target != NF_ACCEPT) {
		printf("--mangle-target ");
		if (m->target == NF_DROP)
			printf("DROP ");
		else
			printf("CONTINUE ");
	}
}

static void
save(const struct arpt_arp *ip, const struct arpt_entry_target *target)
{
	struct arpt_mangle *m = (struct arpt_mangle *)(target->data);

	if (m->flags & ARPT_MANGLE_SIP) {
		printf("--mangle-ip-s %s ", addr_to_dotted(&(m->u_s.src_ip)) );
	}
	if (m->flags & ARPT_MANGLE_SDEV) {
		printf("--mangle-mac-s ");
		print_hwaddr(m->src_devaddr, 6);
		printf(" ");
	}
	if (m->flags & ARPT_MANGLE_TIP) {
		printf("--mangle-ip-d %s ", addr_to_dotted(&(m->u_t.tgt_ip)) );
	}
	if (m->flags & ARPT_MANGLE_TDEV) {
		printf("--mangle-mac-d ");
		print_hwaddr(m->tgt_devaddr, 6);
		printf(" ");
	}
	if (m->target != NF_ACCEPT) {
		printf("--mangle-target ");
		if (m->target == NF_DROP)
			printf("DROP ");
		else
			printf("CONTINUE ");
	}
}

static
struct arptables_target change
= { NULL,
    "mangle",
    ARPTABLES_VERSION,
    ARPT_ALIGN(sizeof(struct arpt_mangle)),
    ARPT_ALIGN(sizeof(struct arpt_mangle)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

static void _init(void) __attribute__ ((constructor));
static void _init(void)
{
	register_target(&change);
}
